# Built-ins
from __future__ import absolute_import, division, print_function
import pickle

# 3rd party
import pytest

# This module
import iteration_utilities

# Test helper
from helper_leak import memory_leak_decorator


complement = iteration_utilities.complement


@memory_leak_decorator()
def test_complement_normal1():
    assert not complement(lambda x: x is True)(True)


@memory_leak_decorator()
def test_complement_normal2():
    assert complement(lambda x: x is True)(False)


@memory_leak_decorator()
def test_complement_normal3():
    assert complement(lambda x: x is False)(True)


@memory_leak_decorator()
def test_complement_normal4():
    assert not complement(lambda x: x is False)(False)


@memory_leak_decorator()
def test_complement_normal5():
    assert not complement(iteration_utilities.is_None)(None)


@memory_leak_decorator()
def test_complement_normal6():
    assert complement(iteration_utilities.is_None)(False)


@memory_leak_decorator()
def test_complement_normal7():
    assert complement(iteration_utilities.is_None)(True)


@memory_leak_decorator(offset=1)
def test_complement_pickle1():
    x = pickle.dumps(complement(iteration_utilities.is_None))
    assert pickle.loads(x)(False)
    assert pickle.loads(x)(True)
    assert not pickle.loads(x)(None)



# Built-ins
from __future__ import absolute_import, division, print_function
import pickle

# 3rd party
import pytest

# This module
import iteration_utilities

# Test helper
from helper_leak import memory_leak_decorator
from helper_cls import T, toT


nth = iteration_utilities.nth


@memory_leak_decorator()
def test_nth_normal1():
    assert nth(1)([T(1), T(2), T(3)]) == T(2)


@memory_leak_decorator()
def test_nth_normal2():
    assert nth(2)(map(T, range(10))) == T(2)


@memory_leak_decorator()
def test_nth_pred1():
    # With pred
    assert nth(1)([T(0), T(1), T(2)], pred=bool) == T(2)


@memory_leak_decorator()
def test_nth_pred2():
    assert nth(1)([T(0), T(1), T(2)], pred=None) == T(2)


@memory_leak_decorator()
def test_nth_pred3():
    assert nth(0)([T(0)]*100 + [T(1)], pred=bool) == T(1)


@memory_leak_decorator()
def test_nth_pred4():
    assert nth(1)([[T(0)], [T(1), T(2)]]*2,
                  pred=lambda x: len(x) > 1) == [T(1), T(2)]


@memory_leak_decorator()
def test_nth_predtruthyretpred1():
    # pred with truthy/retpred
    assert nth(1)([T(0), T(2), T(3), T(0)], pred=bool, truthy=False) == T(0)


@memory_leak_decorator()
def test_nth_predtruthyretpred2():
    assert not nth(1)([T(0), T(1), T(2), T(3), T(0)], pred=bool, truthy=False,
                      retpred=True)


@memory_leak_decorator()
def test_nth_predtruthyretpred3():
    assert nth(1)([T(0), T(2), T(3), T(0)], pred=lambda x: x**T(2),
                  truthy=False) == T(0)


@memory_leak_decorator()
def test_nth_predtruthyretpred4():
    assert nth(1)(toT([0, 1, 2, 3, 0]),
                  pred=lambda x: x**T(2), truthy=False, retpred=True) == T(0)


@memory_leak_decorator()
def test_nth_predtruthyretpred5():
    assert nth(2)([T(0), T(1), T(2), T(3)], pred=bool) == T(3)


@memory_leak_decorator()
def test_nth_predtruthyretpred6():
    assert nth(2)([T(0), T(1), T(2), T(3)], pred=bool, retpred=True)


@memory_leak_decorator()
def test_nth_predtruthyretpred7():
    assert nth(2)([T(0), T(1), T(2), T(3)], pred=lambda x: x**T(2)) == T(3)


@memory_leak_decorator()
def test_nth_predtruthyretpred8():
    assert nth(2)([T(0), T(2), T(3), T(4)], pred=lambda x: x**T(2),
                  retpred=True) == T(16)


@memory_leak_decorator()
def test_nth_default1():
    # With default
    assert nth(2)([], default=None) is None


@memory_leak_decorator()
def test_nth_default2():
    assert nth(1)([T(0), T(0), T(0)], default=None, pred=bool) is None


@memory_leak_decorator(collect=True)
def test_nth_failures1():
    # failures
    with pytest.raises(TypeError):
        nth(10)(T(100))


@memory_leak_decorator(collect=True)
def test_nth_failures2():
    with pytest.raises(IndexError):
        nth(10)([])


@memory_leak_decorator(collect=True)
def test_nth_failures3():
    with pytest.raises(IndexError):
        nth(1)([T(0)], pred=bool)


@memory_leak_decorator(collect=True)
def test_nth_failures4():
    with pytest.raises(TypeError):
        nth(1)([T('a'), T('b')], pred=lambda x: abs(x.value))


@memory_leak_decorator(collect=True)
def test_nth_failures5():
    with pytest.raises(TypeError):
        nth('a')


@memory_leak_decorator(offset=1)
def test_nth_pickle1():
    x = pickle.dumps(nth(2))
    assert pickle.loads(x)([T(1), T(2), T(3), T(4)]) == T(3)



# Built-ins
from __future__ import absolute_import, division, print_function
import pickle

# 3rd party
import pytest

# This module
import iteration_utilities

# Test helper
from helper_leak import memory_leak_decorator
from helper_cls import T


one = iteration_utilities.one


@memory_leak_decorator()
def test_one_normal1():
    assert one([T(0)]) == T(0)


@memory_leak_decorator()
def test_one_normal2():
    assert one('a') == 'a'


@memory_leak_decorator()
def test_one_normal3():
    assert one({T('o'): T(10)}) == T('o')


@memory_leak_decorator(collect=True)
def test_one_failure1():
    with pytest.raises(TypeError):
        one(T(0))


@memory_leak_decorator(collect=True)
def test_one_failure2():
    # empty iterable
    with pytest.raises(ValueError):
        one([])


@memory_leak_decorator(collect=True)
def test_one_failure3():
    # more than 1 element
    with pytest.raises(ValueError):
        one([T(1), T(2)])



# Built-ins
from __future__ import absolute_import, division, print_function
import operator
import pickle

# 3rd party
import pytest

# This module
import iteration_utilities

# Test helper
from helper_leak import memory_leak_decorator
from helper_cls import T


accumulate = iteration_utilities.accumulate


@memory_leak_decorator()
def test_accumulate_empty1():
    assert list(accumulate([])) == []


@memory_leak_decorator()
def test_accumulate_normal1():
    assert list(accumulate([T(1), T(2), T(3)])) == [T(1), T(3), T(6)]


@memory_leak_decorator()
def test_accumulate_normal2():
    assert list(accumulate([], None)) == []


@memory_leak_decorator()
def test_accumulate_normal3():
    assert list(accumulate([T(1), T(2), T(3), T(4)],
                           None)) == [T(1), T(3), T(6), T(10)]


@memory_leak_decorator()
def test_accumulate_binop1():
    assert list(accumulate([T(1), T(2), T(3), T(4)],
                           operator.add)) == [T(1), T(3), T(6), T(10)]


@memory_leak_decorator()
def test_accumulate_binop2():
    assert list(accumulate([T(1), T(2), T(3), T(4)],
                           operator.mul)) == [T(1), T(2), T(6), T(24)]


@memory_leak_decorator()
def test_accumulate_initial1():
    assert list(accumulate([T(1), T(2), T(3)],
                           None, T(10))) == [T(11), T(13), T(16)]


@memory_leak_decorator(collect=True)
def test_accumulate_failure1():
    with pytest.raises(TypeError):
        list(accumulate([T(1), T(2), T(3)], None, T('a')))


@memory_leak_decorator(collect=True)
def test_accumulate_failure2():
    with pytest.raises(TypeError):
        list(accumulate([T(1), T(2), T(3)], operator.add, T('a')))


@memory_leak_decorator(collect=True)
def test_accumulate_failure3():
    with pytest.raises(TypeError):
        list(accumulate([T('a'), T(2), T(3)]))


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_accumulate_pickle1():
    acc = accumulate([T(1), T(2), T(3), T(4)])
    assert next(acc) == T(1)
    x = pickle.dumps(acc)
    assert list(pickle.loads(x)) == [T(3), T(6), T(10)]


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_accumulate_pickle2():
    acc = accumulate([T(1), T(2), T(3), T(4)])
    x = pickle.dumps(acc)
    assert list(pickle.loads(x)) == [T(1), T(3), T(6), T(10)]


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_accumulate_pickle3():
    acc = accumulate([T(1), T(2), T(3), T(4)], operator.mul)
    assert next(acc) == T(1)
    x = pickle.dumps(acc)
    assert list(pickle.loads(x)) == [T(2), T(6), T(24)]


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_accumulate_pickle4():
    acc = accumulate([T(1), T(2), T(3), T(4)], None, T(4))
    x = pickle.dumps(acc)
    assert list(pickle.loads(x)) == [T(5), T(7), T(10), T(14)]



# Built-ins
from __future__ import absolute_import, division, print_function
import operator
import pickle

# 3rd party
import pytest

# This module
import iteration_utilities

# Test helper
from helper_leak import memory_leak_decorator
from helper_cls import T


unique_everseen = iteration_utilities.unique_everseen
Seen = iteration_utilities.Seen


@memory_leak_decorator()
def test_uniqueeverseen_empty1():
    assert list(unique_everseen([])) == []


@memory_leak_decorator()
def test_uniqueeverseen_normal1():
    assert list(unique_everseen([T(1), T(2), T(1)])) == [T(1), T(2)]


@memory_leak_decorator()
def test_uniqueeverseen_key1():
    assert list(unique_everseen([T(1), T(2), T(1)], abs)) == [T(1), T(2)]


@memory_leak_decorator()
def test_uniqueeverseen_key2():
    assert list(unique_everseen([T(1), T(1), T(-1)], abs)) == [T(1)]


@memory_leak_decorator()
def test_uniqueeverseen_unhashable1():
    assert list(unique_everseen([{T(1): T(1)}, {T(2): T(2)},
                                 {T(1): T(1)}])) == [{T(1): T(1)},
                                                     {T(2): T(2)}]


@memory_leak_decorator()
def test_uniqueeverseen_unhashable2():
    assert list(unique_everseen([[T(1)], [T(2)], [T(1)]])) == [[T(1)], [T(2)]]


@memory_leak_decorator()
def test_uniqueeverseen_unhashable3():
    assert list(unique_everseen([[T(1), T(1)], [T(1), T(2)], [T(1), T(3)]],
                                operator.itemgetter(0))) == [[T(1), T(1)]]


@memory_leak_decorator()
def test_uniqueeverseen_getter1():
    t = unique_everseen([T(1), T([0, 0]), T(3)])
    assert not t.seen
    assert t.key is None
    assert next(t) == T(1)
    assert t.seen == Seen({T(1)})
    assert t.key is None
    assert next(t) == T([0, 0])
    assert T(1) in t.seen
    assert T([0, 0]) in t.seen
    assert t.key is None
    assert next(t) == T(3)
    assert t.seen == Seen({T(1), T(3)}, [T([0, 0])])
    assert t.key is None


@memory_leak_decorator(collect=True)
def test_uniqueeverseen_failure1():
    with pytest.raises(TypeError):
        list(unique_everseen(T(10)))


@memory_leak_decorator(collect=True)
def test_uniqueeverseen_failure2():
    with pytest.raises(TypeError):
        list(unique_everseen([T(1), T(2), T(3), T('a')], abs))


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_uniqueeverseen_pickle1():
    uqe = unique_everseen([T(1), T(2), T(1), T(2)])
    assert next(uqe) == T(1)
    x = pickle.dumps(uqe)
    assert list(pickle.loads(x)) == [T(2)]



# Built-ins
from __future__ import absolute_import, division, print_function
import pickle

# 3rd party
import pytest

# This module
import iteration_utilities

# Test helper
from helper_leak import memory_leak_decorator
from helper_cls import T as Original_T


chained = iteration_utilities.chained


class T(Original_T):
    def __add__(self, other):
        return self.__class__(self.value + other)

    def __mul__(self, other):
        return self.__class__(self.value * other)

    def __rtruediv__(self, other):
        return self.__class__(other / self.value)

    def __pow__(self, other):
        return self.__class__(self.value ** other)


@memory_leak_decorator()
def test_chained_normal1():
    double_increment = chained(lambda x: x*2, lambda x: x+1)
    assert double_increment(T(10)) == T(21)
    assert double_increment(T(2)) == T(5)


@memory_leak_decorator()
def test_chained_reverse1():
    double_increment = chained(lambda x: x*2, lambda x: x+1, reverse=True)
    assert double_increment(T(10)) == T(22)
    assert double_increment(T(2)) == T(6)


@memory_leak_decorator()
def test_chained_all1():
    double_increment = chained(lambda x: x*2, lambda x: x+1, all=True)
    assert double_increment(T(10)) == (T(20), T(11))
    assert double_increment(T(2)) == (T(4), T(3))


@memory_leak_decorator(collect=True)
def test_chained_failure1():
    with pytest.raises(TypeError):  # at least one func must be present
        chained()


@memory_leak_decorator(collect=True)
def test_chained_failure2():
    with pytest.raises(TypeError):  # kwarg not accepted
        chained(lambda x: x+1, invalidkwarg=lambda x: x*2)


@memory_leak_decorator(collect=True)
def test_chained_failure3():
    with pytest.raises(TypeError):  # func fails
        chained(lambda x: x+1)(T('a'))


@memory_leak_decorator(collect=True)
def test_chained_failure4():
    with pytest.raises(TypeError):  # second func fails
        chained(lambda x: x*2, lambda x: x+1)(T('a'))


@memory_leak_decorator(offset=1)
def test_chained_pickle1():
    cmp = chained(iteration_utilities.square, iteration_utilities.reciprocal)
    x = pickle.dumps(cmp)
    assert pickle.loads(x)(T(10)) == T(1/100)
    assert pickle.loads(x)(T(2)) == T(1/4)


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_chained_pickle2():
    cmp = chained(iteration_utilities.square, iteration_utilities.double,
                  reverse=True)
    x = pickle.dumps(cmp)
    assert pickle.loads(x)(T(10)) == T(400)
    assert pickle.loads(x)(T(3)) == T(36)


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_chained_pickle3():
    cmp = chained(iteration_utilities.square, iteration_utilities.double,
                  all=True)
    x = pickle.dumps(cmp)
    assert pickle.loads(x)(T(10)) == (T(100), T(20))
    assert pickle.loads(x)(T(3)) == (T(9), T(6))



# Built-ins
from __future__ import absolute_import, division, print_function

# 3rd party
import pytest

# This module
import iteration_utilities

# Test helper
from helper_leak import memory_leak_decorator
from helper_cls import T, toT


partition = iteration_utilities.partition


@memory_leak_decorator()
def test_partition_empty1():
    assert partition([]) == (toT([]), toT([]))


@memory_leak_decorator()
def test_partition_normal1():
    assert partition([T(0), T(1), T(2)]) == (toT([0]), toT([1, 2]))


@memory_leak_decorator()
def test_partition_normal2():
    assert partition([T(3), T(1), T(0)]) == (toT([0]), toT([3, 1]))


@memory_leak_decorator()
def test_partition_normal3():
    assert partition([T(0), T(0), T(0)]) == (toT([0, 0, 0]), [])


@memory_leak_decorator()
def test_partition_normal4():
    assert partition([T(1), T(1), T(1)]) == ([], toT([1, 1, 1]))


@memory_leak_decorator()
def test_partition_pred1():
    assert partition([T(0), T(1), T(2)],
                     lambda x: x.value > 1) == (toT([0, 1]), toT([2]))


@memory_leak_decorator()
def test_partition_pred2():
    assert partition([T(0), T(1), T(2)],
                     lambda x: x.value < 1) == (toT([1, 2]), toT([0]))


@memory_leak_decorator(collect=True)
def test_partition_failure1():
    # not-iterable
    with pytest.raises(TypeError):
        partition(T(10))


@memory_leak_decorator(collect=True)
def test_partition_failure2():
    with pytest.raises(TypeError):
        partition([T(1), T('a')], lambda x: x.value + 3)


@memory_leak_decorator(collect=True)
def test_partition_failure3():

    with pytest.raises(TypeError):
        partition([T(1), T('a')], lambda x: x.value - 1)


@memory_leak_decorator(collect=True)
def test_partition_failure4():
    with pytest.raises(TypeError):
        partition([T(1), T('a')], lambda x: x.value + 'a')



# Built-ins
from __future__ import absolute_import, division, print_function
import pickle

# 3rd party
import pytest

# This module
import iteration_utilities

# Test helper
from helper_leak import memory_leak_decorator
from helper_cls import T, toT


clamp = iteration_utilities.clamp


@memory_leak_decorator()
def test_applyfunc_empty1():
    assert list(clamp([], T(10), T(100))) == []


@memory_leak_decorator()
def test_applyfunc_normal1():
    assert list(clamp(toT(range(10)), T(2), T(7))) == toT([2, 3, 4, 5, 6, 7])


@memory_leak_decorator()
def test_applyfunc_normal2():
    # only low
    assert list(clamp(toT(range(10)), T(2))) == toT([2, 3, 4, 5, 6, 7, 8, 9])


@memory_leak_decorator()
def test_applyfunc_normal3():
    # only high
    assert list(clamp(toT(range(10)),
                      high=T(7))) == toT([0, 1, 2, 3, 4, 5, 6, 7])


@memory_leak_decorator()
def test_applyfunc_normal4():
    # both, inclusive
    assert list(clamp(toT(range(10)), low=T(2), high=T(7),
                      inclusive=True)) == toT([3, 4, 5, 6])


@memory_leak_decorator()
def test_applyfunc_normal5():
    # no low/high
    assert list(clamp(toT(range(10)))) == toT([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])


@pytest.mark.xfail(iteration_utilities.PY2, reason='cmp works on Python 2')
@memory_leak_decorator(collect=True)
def test_applyfunc_failure1():
    with pytest.raises(TypeError):
        list(clamp(toT(range(10)), T('a'), T(3)))


@pytest.mark.xfail(iteration_utilities.PY2, reason='cmp works on Python 2')
@memory_leak_decorator(collect=True)
def test_applyfunc_failure2():
    with pytest.raises(TypeError):
        list(clamp(map(T, range(10)), T(3), T('a')))


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_applyfunc_pickle1():
    clmp = clamp(toT(range(10)), T(2), T(7))
    assert next(clmp) == T(2)
    x = pickle.dumps(clmp)
    assert list(pickle.loads(x)) == toT([3, 4, 5, 6, 7])


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_applyfunc_pickle2():
    # inclusive
    clmp = clamp(map(T, range(10)), T(2), T(7), inclusive=True)
    assert next(clmp) == T(3)
    x = pickle.dumps(clmp)
    assert list(pickle.loads(x)) == toT([4, 5, 6])


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_applyfunc_pickle3():
    # only low
    clmp = clamp(map(T, range(10)), T(2))
    assert next(clmp) == T(2)
    x = pickle.dumps(clmp)
    assert list(pickle.loads(x)) == toT([3, 4, 5, 6, 7, 8, 9])


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_applyfunc_pickle4():
    # only high
    clmp = clamp(map(T, range(10)), high=T(7))
    assert next(clmp) == T(0)
    x = pickle.dumps(clmp)
    assert list(pickle.loads(x)) == toT([1, 2, 3, 4, 5, 6, 7])


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_applyfunc_pickle5():
    # only high, with inclusive
    clmp = clamp(map(T, range(10)), high=T(7), inclusive=True)
    assert next(clmp) == T(0)
    x = pickle.dumps(clmp)
    assert list(pickle.loads(x)) == toT([1, 2, 3, 4, 5, 6])


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_applyfunc_pickle6():
    # only low, with inclusive
    clmp = clamp(map(T, range(10)), T(2), inclusive=True)
    assert next(clmp) == T(3)
    x = pickle.dumps(clmp)
    assert list(pickle.loads(x)) == toT([4, 5, 6, 7, 8, 9])


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_applyfunc_pickle7():
    # no low no high
    clmp = clamp(map(T, range(10)))
    assert next(clmp) == T(0)
    x = pickle.dumps(clmp)
    assert list(pickle.loads(x)) == toT([1, 2, 3, 4, 5, 6, 7, 8, 9])



# Built-ins
from __future__ import absolute_import, division, print_function
import pickle

# 3rd party
import pytest

# This module
import iteration_utilities

# Test helper
from helper_leak import memory_leak_decorator
from helper_cls import T


ItemIdxKey = iteration_utilities.ItemIdxKey


@memory_leak_decorator()
def test_itemidxkey_getter():
    iik = ItemIdxKey(T(10), 2)
    assert iik.item == T(10)
    assert iik.idx == 2
    assert iik.key is None

    iik = ItemIdxKey(T(10), 2, T(5))
    assert iik.item == T(10)
    assert iik.idx == 2
    assert iik.key == T(5)


@memory_leak_decorator()
def test_itemidxkey_setter():
    iik = ItemIdxKey(T(10), 2)
    iik.item = T(20)
    assert iik.item == T(20)
    iik.idx = 10
    assert iik.idx == 10
    iik.key = T(0)
    assert iik.key == T(0)

    iik = ItemIdxKey(T(10), 2, T(5))
    iik.item = T(20)
    assert iik.item == T(20)
    iik.idx = 10
    assert iik.idx == 10
    iik.key = T(0)
    assert iik.key == T(0)


@memory_leak_decorator(collect=True)
def test_itemidxkey_setter_failure():
    iik = ItemIdxKey(T(10), 2)
    with pytest.raises(TypeError):
        iik.idx = 'a'

    iik = ItemIdxKey(T(10), 2, T(5))
    with pytest.raises(TypeError):
        iik.idx = 'a'


@memory_leak_decorator()
def test_itemidxkey_deleter():
    iik = ItemIdxKey(T(10), 2)
    del iik.key
    assert iik.key is None

    iik = ItemIdxKey(T(10), 2, T(5))
    del iik.key
    assert iik.key is None


@memory_leak_decorator(collect=True)
def test_itemidxkey_deleter_failure():
    iik = ItemIdxKey(T(10), 2)
    with pytest.raises(TypeError):
        del iik.item
    with pytest.raises(TypeError):
        del iik.idx

    iik = ItemIdxKey(T(10), 2, T(5))
    with pytest.raises(TypeError):
        del iik.item
    with pytest.raises(TypeError):
        del iik.idx


@memory_leak_decorator(offset=1)
def test_itemidxkey_pickle1():
    iik = ItemIdxKey(T(10), 2)
    x = pickle.dumps(iik)
    assert pickle.loads(x).item == T(10)
    assert pickle.loads(x).idx == 2


@memory_leak_decorator(offset=1)
def test_itemidxkey_pickle2():
    iik = ItemIdxKey(T(10), 2, T(5))
    x = pickle.dumps(iik)
    assert pickle.loads(x).item == T(10)
    assert pickle.loads(x).idx == 2
    assert pickle.loads(x).key == T(5)



# Built-ins
from __future__ import absolute_import, division, print_function
import collections
import gc
import weakref

# This module
import iteration_utilities

if iteration_utilities.PY2:
    range = xrange


def _make_tuple(obj):
    if obj is None:
        return tuple()
    if not isinstance(obj, tuple):
        return (obj, )
    else:
        return obj


def memory_leak_decorator(specific_object=None, exclude_object=weakref.ref,
                          collect=False, offset=0):
    """Compares the number of tracked python objects before and after a
    function call and returns a dict containing differences.

    Parameters
    ----------
    specific_object : type, tuple of types, None, optional
        Test all tracked types (if it's ``None``) or only the specific type(s).
        Default is ``None``.

    exclude_object : type, tuple of types, None, optional
        Exclude specific type(s) or use all (if it's ``None``).
        Default is ``weakref.ref``.

    collect : bool, optional
        If ``True`` remove cyclic references by calling ``gc.collect`` after
        the function call.
        Default is ``False``.

    offset : int, optional
        Call the function `offset` times before doing the check for a memory
        leak. If the function is unpure (relies on external states) calls might
        introduce some "wanted" memory leaks.
        Default is ``0``.

    Returns
    -------
    decorator : callable, optional
        Returns a decorator which can be applied to test functions.

    Raises
    ------
    TypeError
        The decoratored function raises this Exception if any Python object
        leaked.

    Notes
    -----
    It is convenient to wrap the explicit call inside a function that doesn't
    return. To enhance this useage the ``memory_leak`` function doesn't allow
    to pass arguments to the ``func``!
    """
    def decorator_factory(func):
        def inner():
            for i in range(offset):
                func()

            # Create Counter before listing the objects otherwise they would
            # be recognized as leak.
            before = collections.Counter()
            after = collections.Counter()

            # Tracked objects before the function call
            before.update(map(type, gc.get_objects()))

            func()
            if collect:
                gc.collect()  # Takes care of cyclic references!

            # Tracked objects after the function call
            after.update(map(type, gc.get_objects()))

            specifics = _make_tuple(specific_object)
            if not specifics:
                specifics = set(after)
            excludes = set(_make_tuple(exclude_object))

            if any(after[specific] - before[specific] > 0
                   for specific in specifics
                   if specific not in excludes and
                   specific in before):
                raise TypeError('leaked objects: {}'.format(
                    {specific: after[specific] - before[specific]
                     for specific in specifics
                     if specific not in excludes and
                     after[specific] - before[specific] > 0}))
        return inner
    return decorator_factory



# Built-ins
from __future__ import absolute_import, division, print_function

# 3rd party
import pytest

# This module
import iteration_utilities

# Test helper
from helper_leak import memory_leak_decorator
from helper_cls import T


argmin = iteration_utilities.argmin
argmax = iteration_utilities.argmax


@memory_leak_decorator()
def test_argmax_normal1():
    # Just one test for argmax because the internals are identical to argmin
    assert argmax(T(0), T(1), T(2)) == 2


@memory_leak_decorator()
def test_argmin_positional1():
    assert argmin(T(0), T(1), T(2)) == 0


@memory_leak_decorator()
def test_argmin_positional2():
    assert argmin(T(3), T(0), T(1)) == 1


@memory_leak_decorator()
def test_argmin_sequence1():
    assert argmin([T(3), T(0), T(1)]) == 1


@memory_leak_decorator()
def test_argmin_generator1():
    assert argmin((T(i) for i in [5, 4, 3, 2])) == 3


@memory_leak_decorator()
def test_argmin_key1():
    assert argmin([T(3), T(-2), T(1)], key=abs) == 2


@memory_leak_decorator()
def test_argmin_default1():
    assert argmin([], default=2) == 2


@memory_leak_decorator(collect=True)
def test_argmin_failure1():
    # default not possible if given multiple positional arguments
    with pytest.raises(TypeError):
        argmin(T(3), T(0), T(1), default=1)


@memory_leak_decorator(collect=True)
def test_argmin_failure2():
    # not integer default value
    with pytest.raises(TypeError):
        argmin([T(3), T(0), T(1)], default='1.5', key=lambda x: x + 1)


@memory_leak_decorator(collect=True)
def test_argmin_failure3():
    # unwanted kwarg
    with pytest.raises(TypeError):
        argmin([T(3), T(0), T(1)], default=1, key=lambda x: x + 1, blub=10)


@memory_leak_decorator(collect=True)
def test_argmin_failure4():
    # no args
    with pytest.raises(TypeError):
        argmin(key=lambda x: x + 1)


@memory_leak_decorator(collect=True)
def test_argmin_failure5():
    # empty sequence
    with pytest.raises(ValueError):
        argmin([], key=lambda x: x + 1)


@pytest.mark.xfail(iteration_utilities.PY2, reason='cmp works on Python 2')
@memory_leak_decorator(collect=True)
def test_argmin_failure6():
    # cmp failed
    with pytest.raises(TypeError):
        argmin([T(1), T(2), T('a')], key=lambda x: x)


@memory_leak_decorator(collect=True)
def test_argmin_failure7():
    # key failed
    with pytest.raises(TypeError):
        argmin([T(1), T(2), T('a')], key=lambda x: x + 1)



# Built-ins
from __future__ import absolute_import, division, print_function

# 3rd party
import pytest

# This module
import iteration_utilities

# Test helper
from helper_leak import memory_leak_decorator
from helper_cls import T


all_monotone = iteration_utilities.all_monotone


@memory_leak_decorator()
def test_all_monotone_empty1():
    assert all_monotone([])


@memory_leak_decorator()
def test_all_monotone_normal1():
    assert all_monotone([T(1), T(1), T(1)])


@memory_leak_decorator()
def test_all_monotone_normal2():
    assert not all_monotone([T(1), T(1), T(1)], strict=True)


@memory_leak_decorator()
def test_all_monotone_normal3():
    assert all_monotone([T(1), T(2), T(3)])


@memory_leak_decorator()
def test_all_monotone_normal4():
    assert all_monotone([T(1), T(2), T(3)], strict=True)


@memory_leak_decorator()
def test_all_monotone_normal5():
    assert all_monotone([T(1), T(1), T(1)], decreasing=True)


@memory_leak_decorator()
def test_all_monotone_normal6():
    assert not all_monotone([T(1), T(1), T(1)], decreasing=True, strict=True)


@memory_leak_decorator()
def test_all_monotone_normal7():
    assert all_monotone([T(3), T(2), T(1)], decreasing=True)


@memory_leak_decorator()
def test_all_monotone_normal8():
    assert all_monotone([T(3), T(2), T(1)], decreasing=True, strict=True)


@memory_leak_decorator(collect=True)
def test_all_monotone_failure1():
    # not iterable
    with pytest.raises(TypeError):
        all_monotone(T(1))


@memory_leak_decorator(collect=True)
def test_all_monotone_failure2():
    # comparison fail
    with pytest.raises(TypeError):
        all_monotone([T(1), T('a')])



# Built-ins
from __future__ import absolute_import, division, print_function
import pickle

# 3rd party
import pytest

# This module
import iteration_utilities

# Test helper
from helper_leak import memory_leak_decorator
from helper_cls import T, toT


grouper = iteration_utilities.grouper


@memory_leak_decorator()
def test_grouper_empty1():
    # Empty iterable
    assert list(grouper([], 2)) == []


@memory_leak_decorator()
def test_grouper_normal1():
    # no fillvalue + truncate
    assert list(grouper([T(1)], 3)) == [(T(1), )]


@memory_leak_decorator()
def test_grouper_normal2():
    assert list(grouper([T(1), T(2)], 3)) == [(T(1), T(2))]


@memory_leak_decorator()
def test_grouper_normal3():
    assert list(grouper([T(1), T(2), T(3)], 3)) == [(T(1), T(2), T(3))]


@memory_leak_decorator()
def test_grouper_normal4():
    assert list(grouper([T(1), T(2), T(3), T(4)], 3)) == [(T(1), T(2), T(3)),
                                                          (T(4), )]


@memory_leak_decorator()
def test_grouper_normal5():
    assert list(grouper(toT([1, 2, 3, 4, 5]), 3)) == [(T(1), T(2), T(3)),
                                                      (T(4), T(5))]


@memory_leak_decorator()
def test_grouper_normal6():
    assert list(grouper(toT([1, 2, 3, 4, 5, 6]), 3)) == [(T(1), T(2), T(3)),
                                                         (T(4), T(5), T(6))]


@memory_leak_decorator()
def test_grouper_fill1():
    # with fillvalue
    assert list(grouper(toT([1]), 3,
                        fillvalue=T(0))) == [(T(1), T(0), T(0))]


@memory_leak_decorator()
def test_grouper_fill2():
    assert list(grouper(toT([1, 2]), 3,
                        fillvalue=T(0))) == [(T(1), T(2), T(0))]


@memory_leak_decorator()
def test_grouper_fill3():
    assert list(grouper(toT([1, 2, 3]), 3,
                        fillvalue=T(0))) == [(T(1), T(2), T(3))]


@memory_leak_decorator()
def test_grouper_fill4():
    assert list(grouper(toT([1, 2, 3, 4]), 3,
                        fillvalue=T(0))) == [(T(1), T(2), T(3)),
                                             (T(4), T(0), T(0))]


@memory_leak_decorator()
def test_grouper_fill5():
    assert list(grouper(toT([1, 2, 3, 4, 5]), 3,
                        fillvalue=T(0))) == [(T(1), T(2), T(3)),
                                             (T(4), T(5), T(0))]


@memory_leak_decorator()
def test_grouper_fill6():
    assert list(grouper(toT([1, 2, 3, 4, 5, 6]), 3,
                        fillvalue=T(0))) == [(T(1), T(2), T(3)),
                                             (T(4), T(5), T(6))]


@memory_leak_decorator()
def test_grouper_truncate1():
    # with truncate
    assert list(grouper(toT([1]), 3, truncate=True)) == []


@memory_leak_decorator()
def test_grouper_truncate2():
    assert list(grouper(toT([1, 2]), 3, truncate=True)) == []


@memory_leak_decorator()
def test_grouper_truncate3():
    assert list(grouper(toT([1, 2, 3]), 3,
                        truncate=True)) == [(T(1), T(2), T(3))]


@memory_leak_decorator()
def test_grouper_truncate4():
    assert list(grouper(toT([1, 2, 3, 4]), 3,
                        truncate=True)) == [(T(1), T(2), T(3))]


@memory_leak_decorator()
def test_grouper_truncate5():
    assert list(grouper(toT([1, 2, 3, 4, 5]), 3,
                        truncate=True)) == [(T(1), T(2), T(3))]


@memory_leak_decorator()
def test_grouper_truncate6():
    assert list(grouper(toT([1, 2, 3, 4, 5, 6]), 3,
                        truncate=True)) == [(T(1), T(2), T(3)),
                                            (T(4), T(5), T(6))]


@memory_leak_decorator(collect=True)
def test_grouper_failure1():
    # fillvalue + truncate is forbidden
    with pytest.raises(TypeError):
        grouper(toT([1, 2, 3]), 2, fillvalue=T(0), truncate=True)


@memory_leak_decorator(collect=True)
def test_grouper_failure2():
    # n must be > 0
    with pytest.raises(ValueError):
        grouper(toT([1, 2, 3]), 0)


@memory_leak_decorator(collect=True)
def test_grouper_failure3():
    # iterable must be iterable
    with pytest.raises(TypeError):
        grouper(T(1), 2)


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_grouper_pickle1():
    grp = grouper(toT(range(10)), 3)
    assert next(grp) == (T(0), T(1), T(2))
    x = pickle.dumps(grp)
    assert list(pickle.loads(x)) == [(T(3), T(4), T(5)), (T(6), T(7), T(8)),
                                     (T(9), )]


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_grouper_pickle2():
    grp = grouper(toT(range(10)), 3, fillvalue=T(0))
    assert next(grp) == (T(0), T(1), T(2))
    x = pickle.dumps(grp)
    assert list(pickle.loads(x)) == [(T(3), T(4), T(5)), (T(6), T(7), T(8)),
                                     (T(9), T(0), T(0))]


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_grouper_pickle3():
    grp = grouper(toT(range(10)), 3, truncate=True)
    assert next(grp) == (T(0), T(1), T(2))
    x = pickle.dumps(grp)
    assert list(pickle.loads(x)) == [(T(3), T(4), T(5)), (T(6), T(7), T(8))]



# Built-ins
from __future__ import absolute_import, division, print_function
import pickle

# 3rd party
import pytest

# This module
import iteration_utilities

# Test helper
from helper_leak import memory_leak_decorator
from helper_cls import T, toT


roundrobin = iteration_utilities.roundrobin


@memory_leak_decorator()
def test_roundrobin_empty1():
    assert list(roundrobin()) == []


@memory_leak_decorator()
def test_roundrobin_empty2():
    assert list(roundrobin([])) == []


@memory_leak_decorator()
def test_roundrobin_empty3():
    assert list(roundrobin([], (), {})) == []


@memory_leak_decorator()
def test_roundrobin_normal1():
    assert list(roundrobin([T(1)], [T(1), T(2)], [T(1), T(2), T(3)]
                           )) == toT([1, 1, 1, 2, 2, 3])


@memory_leak_decorator()
def test_roundrobin_normal2():
    assert list(roundrobin([T(1), T(2), T(3)], [T(1)], [T(1), T(2)]
                           )) == toT([1, 1, 1, 2, 2, 3])


@memory_leak_decorator()
def test_roundrobin_normal3():
    assert list(roundrobin([T(1), T(2)], [T(1), T(2), T(3)], [T(1)]
                           )) == toT([1, 1, 1, 2, 2, 3])


@memory_leak_decorator(collect=True)
def test_roundrobin_failure1():
    with pytest.raises(TypeError):
        list(roundrobin(T(1)))


@memory_leak_decorator(collect=True)
def test_roundrobin_failure2():
    with pytest.raises(TypeError):
        list(roundrobin([T(1)], T(1)))


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_roundrobin_pickle1():
    rr = roundrobin([T(1), T(2), T(3)], [T(1), T(2), T(3)])
    assert next(rr) == T(1)
    x = pickle.dumps(rr)
    assert list(pickle.loads(x)) == toT([1, 2, 2, 3, 3])


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_roundrobin_pickle2():
    rr2 = roundrobin([T(1)], [T(1), T(2), T(3)])
    assert next(rr2) == T(1)
    assert next(rr2) == T(1)
    assert next(rr2) == T(2)
    x = pickle.dumps(rr2)
    assert list(pickle.loads(x)) == [T(3)]



# Built-ins
from __future__ import absolute_import, division, print_function
import pickle

# 3rd party
import pytest

# This module
import iteration_utilities

# Test helper
from helper_leak import memory_leak_decorator
from helper_cls import T, toT


intersperse = iteration_utilities.intersperse


@memory_leak_decorator()
def test_intersperse_empty1():
    assert list(intersperse([], T(0))) == []


@memory_leak_decorator()
def test_intersperse_empty2():
    assert list(intersperse([T(1)], T(0))) == [T(1)]


@memory_leak_decorator()
def test_intersperse_normal1():
    assert list(intersperse([T(1), T(2)], T(0))) == toT([1, 0, 2])


@memory_leak_decorator(collect=True)
def test_intersperse_failure1():
    with pytest.raises(TypeError):
        intersperse(T(100), T(0))


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_intersperse_pickle1():
    its = intersperse(toT([1, 2, 3]), T(0))
    x = pickle.dumps(its)
    assert list(pickle.loads(x)) == toT([1, 0, 2, 0, 3])


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_intersperse_pickle2():
    its = intersperse(toT([1, 2, 3]), T(0))
    assert next(its) == T(1)
    x = pickle.dumps(its)
    assert list(pickle.loads(x)) == toT([0, 2, 0, 3])


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_intersperse_pickle3():
    its = intersperse([T(1), T(2), T(3)], T(0))
    assert next(its) == T(1)
    assert next(its) == T(0)
    x = pickle.dumps(its)
    assert list(pickle.loads(x)) == toT([2, 0, 3])


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_intersperse_pickle4():
    its = intersperse([T(1), T(2), T(3)], T(0))
    assert next(its) == T(1)
    assert next(its) == T(0)
    assert next(its) == T(2)
    x = pickle.dumps(its)
    assert list(pickle.loads(x)) == toT([0, 3])



# Built-ins
from __future__ import absolute_import, division, print_function
import pickle

# 3rd party
import pytest

# This module
import iteration_utilities

# Test helper
from helper_leak import memory_leak_decorator
from helper_cls import T


successive = iteration_utilities.successive


@memory_leak_decorator()
def test_successive_empty1():
    assert list(successive([])) == []


@memory_leak_decorator()
def test_successive_empty2():
    assert list(successive([T(1)])) == []


@memory_leak_decorator()
def test_successive_empty3():
    assert list(successive([], times=10)) == []


@memory_leak_decorator()
def test_successive_empty4():
    assert list(successive([T(1), T(2), T(3)], times=10)) == []


@memory_leak_decorator()
def test_successive_normal1():
    assert (list(successive([T(1), T(2), T(3), T(4)])) ==
            [(T(1), T(2)), (T(2), T(3)), (T(3), T(4))])


@memory_leak_decorator()
def test_successive_normal2():
    assert (list(successive([T(1), T(2), T(3), T(4)], times=3)) ==
            [(T(1), T(2), T(3)), (T(2), T(3), T(4))])


@memory_leak_decorator()
def test_successive_normal3():
    assert (list(successive([T(1), T(2), T(3), T(4)], times=4)) ==
            [(T(1), T(2), T(3), T(4))])


@memory_leak_decorator()
def test_successive_normal4():
    assert (dict(successive([T(1), T(2), T(3), T(4)])) ==
            {T(1): T(2), T(2): T(3), T(3): T(4)})


@memory_leak_decorator(collect=True)
def test_successive_failure1():
    with pytest.raises(TypeError):
        successive(T(1))


@memory_leak_decorator(collect=True)
def test_successive_failure2():
    with pytest.raises(ValueError):  # times must be > 0
        successive([T(1), T(2), T(3)], 0)


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_successive_pickle1():
    suc = successive([T(1), T(2), T(3), T(4)])
    assert next(suc) == (T(1), T(2))
    x = pickle.dumps(suc)
    assert list(pickle.loads(x)) == [(T(2), T(3)), (T(3), T(4))]


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_successive_pickle2():
    suc = successive([T(1), T(2), T(3), T(4)])
    x = pickle.dumps(suc)
    assert list(pickle.loads(x)) == [(T(1), T(2)), (T(2), T(3)), (T(3), T(4))]



# Built-ins
from __future__ import absolute_import, division, print_function
import pickle

# 3rd party
import pytest

# This module
import iteration_utilities

# Test helper
from helper_leak import memory_leak_decorator
from helper_cls import T, toT


unique_justseen = iteration_utilities.unique_justseen


class T2(object):
    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        raise TypeError()

    def __ne__(self, other):
        raise TypeError()


@memory_leak_decorator()
def test_unique_justseen_empty1():
    assert list(unique_justseen([])) == [] == []


@memory_leak_decorator()
def test_unique_justseen_normal1():
    assert list(unique_justseen(toT([1, 1, 2, 2, 3, 3]))) == toT([1, 2, 3])


@memory_leak_decorator()
def test_unique_justseen_normal2():
    assert list(unique_justseen('aAabBb')) == ['a', 'A', 'a', 'b', 'B', 'b']


@memory_leak_decorator()
def test_unique_justseen_normal3():
    assert list(unique_justseen('aAabBb', key=str.lower)) == ['a', 'b']


@memory_leak_decorator(collect=True)
def test_unique_justseen_failure1():
    # not iterable
    with pytest.raises(TypeError):
        unique_justseen(T(1))


@memory_leak_decorator(collect=True)
def test_unique_justseen_failure2():
    with pytest.raises(TypeError):  # function call fails
        list(unique_justseen([T(1), T(2), T(3)], key=lambda x: x + 'a'))


@memory_leak_decorator(collect=True)
def test_unique_justseen_failure3():
    # objects do not support eq or ne
    with pytest.raises(TypeError):
        list(unique_justseen([T2(1), T2(2)]))


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_unique_justseen_pickle1():
    ujs = unique_justseen([T(1), T(2), T(3)])
    x = pickle.dumps(ujs)
    assert list(pickle.loads(x)) == toT([1, 2, 3])


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_unique_justseen_pickle2():
    ujs = unique_justseen([T(1), T(2), T(3)])
    assert next(ujs) == T(1)
    x = pickle.dumps(ujs)
    assert list(pickle.loads(x)) == toT([2, 3])


@pytest.mark.xfail(not iteration_utilities.PY34, reason='see method comments')
@memory_leak_decorator(offset=1)
def test_unique_justseen_pickle3():
    # Pickling a method descriptor is not possible for Python 3.3 and before
    # Also operator.methodcaller loses it's methodname when pickled for Python
    #   3.4 and lower...
    ujs = unique_justseen(['a', 'A', 'a'], key=str.lower)
    x = pickle.dumps(ujs)
    assert list(pickle.loads(x)) == ['a']


@pytest.mark.xfail(not iteration_utilities.PY34, reason='see method comments')
@memory_leak_decorator(offset=1)
def test_unique_justseen_pickle4():
    # Pickling a method descriptor is not possible for Python 3.3 and before
    # Also operator.methodcaller loses it's methodname when pickled for Python
    #   3.4 and lower...
    ujs = unique_justseen(['a', 'A', 'a'], key=str.lower)
    assert next(ujs) == 'a'
    x = pickle.dumps(ujs)
    assert list(pickle.loads(x)) == []



# Built-ins
from __future__ import absolute_import, division, print_function
import pickle

# 3rd party
import pytest

# This module
import iteration_utilities

# Test helper
from helper_leak import memory_leak_decorator
from helper_cls import T


const = iteration_utilities.constant


@memory_leak_decorator()
def test_constant_normal1():
    one = const(T(1))
    assert one() == T(1)


@memory_leak_decorator()
def test_constant_normal2():
    one = const(T(1))
    assert one(10, a=2) == T(1)


@memory_leak_decorator(offset=1)
def test_constant_pickle1():
    x = pickle.dumps(const(T(10)))
    assert pickle.loads(x)() == T(10)



# Built-ins
from __future__ import absolute_import, division, print_function
import pickle

# 3rd party
import pytest

# This module
import iteration_utilities

# Test helper
from helper_leak import memory_leak_decorator
from helper_cls import T


count_items = iteration_utilities.count_items


@memory_leak_decorator()
def test_count_empty1():
    assert count_items([]) == 0


@memory_leak_decorator()
def test_count_normal1():
    assert count_items([T(0), T(0)]) == 2


@memory_leak_decorator()
def test_count_normal2():
    assert count_items([T(0), T(0), T(1)], bool) == 1


@memory_leak_decorator()
def test_count_normal3():
    assert count_items([T(0), T(0), T(1), T(1)], None) == 4


@memory_leak_decorator()
def test_count_normal4():
    assert count_items([], iteration_utilities.return_identity) == 0


@memory_leak_decorator()
def test_count_normal5():
    assert count_items([T(1), T(2), T(3)], lambda x: x > T(2)) == 1


@memory_leak_decorator()
def test_count_normal6():
    assert count_items([T(1), T(2), T(3)], lambda x: x < T(3)) == 2


@memory_leak_decorator()
def test_count_normal7():
    assert count_items([T(3), T(1), T(2), T(3), T(3)], T(3), True) == 3


@memory_leak_decorator(collect=True)
def test_count_failure1():
    with pytest.raises(TypeError):
        count_items(T(1))


@memory_leak_decorator(collect=True)
def test_count_failure2():
    with pytest.raises(TypeError):
        count_items([T(1)], T(1))



# Built-ins
from __future__ import absolute_import, division, print_function

# 3rd party
import pytest

# This module
import iteration_utilities

# Test helper
from helper_doctest import doctest_module_no_failure


def test_doctests():
    doctest_module_no_failure(iteration_utilities.core)


Iterable = iteration_utilities.Iterable


def test_sentinelfactory():
    as_str = str(iteration_utilities._default)
    as_repr = repr(iteration_utilities._default)
    assert as_str == as_repr
    assert as_str == "<default>"


def test_cls_exception():
    with pytest.raises(TypeError):
        Iterable.from_count().pad(ntail=None)

    # __getitem__ : negative idx
    with pytest.raises(ValueError):
        Iterable(range(10))[-2]

    # __getitem__ : negative step
    with pytest.raises(ValueError):
        Iterable(range(10))[::-2]

    # __getitem__ : positive start with negative stop
    with pytest.raises(ValueError):
        Iterable(range(10))[2:-1]

    # __getitem__ : negative start/stop with infinite iterable.
    with pytest.raises(TypeError):
        Iterable.from_count()[-5:-3]

    # __getitem__ : not int, not slice
    with pytest.raises(TypeError):
        Iterable.from_count()['bad']



# Built-ins
from __future__ import absolute_import, division, print_function
import operator

# 3rd party
import pytest

# This module
import iteration_utilities

# Test helper
from helper_leak import memory_leak_decorator
from helper_cls import T, toT


groupedby = iteration_utilities.groupedby


if iteration_utilities.PY2:
    range = xrange


@memory_leak_decorator()
def test_groupedby_empty1():
    assert groupedby([], key=lambda x: x) == {}


@memory_leak_decorator()
def test_groupedby_normal1():
    assert groupedby([T('a'), T('ab'), T('abc')], key=lambda x: x.value[0]
                     ) == {'a': toT(['a', 'ab', 'abc'])}


@memory_leak_decorator()
def test_groupedby_normal2():
    assert groupedby([T('a'), T('ba'), T('ab'), T('abc'), T('b')],
                     key=lambda x: x.value[0]
                     ) == {'a': toT(['a', 'ab', 'abc']),
                           'b': toT(['ba', 'b'])}


@memory_leak_decorator()
def test_groupedby_keep1():
    assert groupedby([T('a'), T('ba'), T('ab'), T('abc'), T('b')],
                     key=lambda x: x.value[0],
                     keep=len) == {'a': [1, 2, 3], 'b': [2, 1]}


@memory_leak_decorator()
def test_groupedby_reduce1():
    assert groupedby([(T('a'), T(1)), (T('a'), T(2)), (T('b'), T(5))],
                     key=operator.itemgetter(0),
                     keep=operator.itemgetter(1),
                     reduce=operator.add) == {T('a'): T(3), T('b'): T(5)}


@memory_leak_decorator()
def test_groupedby_reduce2():
    assert groupedby([(T('a'), T(1)), (T('a'), T(2)), (T('b'), T(5))],
                     key=operator.itemgetter(0),
                     reduce=lambda x, y: x + y[1],
                     reducestart=T(0)) == {T('a'): T(3), T('b'): T(5)}


@memory_leak_decorator()
def test_groupedby_reduce3():
    assert groupedby(map(T, range(500)), key=lambda x: T(x.value % 5),
                     reduce=operator.add,
                     reducestart=T(0))


@memory_leak_decorator(collect=True)
def test_groupedby_failure1():
    # not iterable
    with pytest.raises(TypeError):
        groupedby(T(1), key=len)


@memory_leak_decorator(collect=True)
def test_groupedby_failure2():
    # key func fails
    with pytest.raises(TypeError):
        groupedby([T(1), T(2), T(3)], key=lambda x: T(x.value + 'a'))


@memory_leak_decorator(collect=True)
def test_groupedby_failure3():
    # keep func fails
    with pytest.raises(TypeError):
        groupedby([T(1), T(2), T(3)], key=lambda x: x,
                  keep=lambda x: T(x.value + 'a'))


@memory_leak_decorator(collect=True)
def test_groupedby_failure4():
    # unhashable
    with pytest.raises(TypeError):
        groupedby([{T('a'): T(10)}], key=lambda x: x)


@memory_leak_decorator(collect=True)
def test_groupedby_failure5():
    # no reduce but reducestart
    with pytest.raises(TypeError):
        groupedby(toT(range(10)), lambda x: x, reducestart=T(0))


@memory_leak_decorator(collect=True)
def test_groupedby_failure6():
    # reduce function fails with reducestart
    with pytest.raises(TypeError):
        groupedby(map(T, range(10)), lambda x: x.value % 2 == 0,
                  reduce=operator.add, reducestart=T('a'))


@memory_leak_decorator(collect=True)
def test_groupedby_failure7():
    # reduce function fails
    with pytest.raises(TypeError):
        groupedby(map(T, [1, 2, 3, 4, 'a']), lambda x: True,
                  reduce=operator.add)



# Built-ins
from __future__ import absolute_import, division, print_function
import operator
import pickle

# 3rd party
import pytest

# This module
import iteration_utilities

# Test helper
from helper_leak import memory_leak_decorator
from helper_cls import T, toT


duplicates = iteration_utilities.duplicates


@memory_leak_decorator()
def test_duplicates_empty1():
    assert list(duplicates([])) == []


@memory_leak_decorator()
def test_duplicates_normal1():
    assert list(duplicates([T(1), T(2), T(1)])) == [T(1)]


@memory_leak_decorator()
def test_duplicates_key1():
    assert list(duplicates([T(1), T(2), T(1)], abs)) == [T(1)]


@memory_leak_decorator()
def test_duplicates_key2():
    assert list(duplicates([T(1), T(1), T(-1)], abs)) == toT([1, -1])


@memory_leak_decorator()
def test_duplicates_unhashable1():
    assert list(duplicates([{T(1): T(1)}, {T(2): T(2)}, {T(1): T(1)}]
                           )) == [{T(1): T(1)}]


@memory_leak_decorator()
def test_duplicates_unhashable2():
    assert list(duplicates([[T(1)], [T(2)], [T(1)]])) == [[T(1)]]


@memory_leak_decorator()
def test_duplicates_unhashable3():
    assert list(duplicates([[T(1), T(1)], [T(1), T(2)],
                            [T(1), T(3)]], operator.itemgetter(0)
                           )) == [[T(1), T(2)], [T(1), T(3)]]


@memory_leak_decorator()
def test_duplicates_getter1():
    t = duplicates([T(1), T([0, 0]), T(3), T(1)])
    assert not t.seen
    assert t.key is None
    assert next(t) == T(1)
    assert T(1) in t.seen
    assert T(3) in t.seen
    assert T([0, 0]) in t.seen
    assert t.key is None


@memory_leak_decorator(collect=True)
def test_duplicates_failure1():
    with pytest.raises(TypeError):
        list(duplicates(T(10)))


@memory_leak_decorator(collect=True)
def test_duplicates_failure2():
    with pytest.raises(TypeError):
        list(duplicates([T(1), T(2), T(3), T('a')], abs))


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_duplicates_pickle1():
    dpl = duplicates([T(1), T(2), T(1), T(2)])
    assert next(dpl) == T(1)
    x = pickle.dumps(dpl)
    assert list(pickle.loads(x)) == [T(2)]



# Built-ins
from __future__ import absolute_import, division, print_function
import operator

# 3rd party
import pytest

# This module
import iteration_utilities

# Test helper
from helper_leak import memory_leak_decorator
from helper_cls import T


minmax = iteration_utilities.minmax


@memory_leak_decorator()
def test_minmax_normal1():
    assert minmax([T(1)]) == (T(1), T(1))


@memory_leak_decorator()
def test_minmax_normal2():
    assert minmax([T(1), T(2)]) == (T(1), T(2))


@memory_leak_decorator()
def test_minmax_normal3():
    assert minmax([T(2), T(1)]) == (T(1), T(2))


@memory_leak_decorator()
def test_minmax_normal4():
    assert minmax([T(1), T(2), T(3)]) == (T(1), T(3))


@memory_leak_decorator()
def test_minmax_normal5():
    assert minmax([T(1), T(3), T(2)]) == (T(1), T(3))


@memory_leak_decorator()
def test_minmax_normal6():
    assert minmax(map(T, range(100))) == (T(0), T(99))


@memory_leak_decorator()
def test_minmax_normal7():
    assert minmax(map(T, range(101))) == (T(0), T(100))


@memory_leak_decorator()
def test_minmax_normal8():
    assert minmax({T(1), T(2), T(-3)}) == (T(-3), T(2))


@memory_leak_decorator()
def test_minmax_normal9():
    assert minmax({T(1): T(0), T(2): T(0), T(3): T(0)}) == (T(1), T(3))


@memory_leak_decorator()
def test_minmax_normal10():
    assert minmax(T(1), T(2), T(3)) == (T(1), T(3))


@memory_leak_decorator()
def test_minmax_normal11():
    assert minmax(T(4), T(3), T(2), T(1)) == (T(1), T(4))


@memory_leak_decorator()
def test_minmax_normal12():
    assert minmax((T(i) for i in [4, 3, 2, 5, 3])) == (T(2), T(5))


@memory_leak_decorator()
def test_minmax_normal13():
    assert minmax((T(i) for i in [4, 3, 2, 5, 3, 3])) == (T(2), T(5))


@memory_leak_decorator()
def test_minmax_normal14():
    assert minmax((T(i) for i in [4, 3, 2, 5, 3]), key=abs) == (T(2), T(5))


@memory_leak_decorator()
def test_minmax_normal15():
    assert minmax((T(i) for i in [4, 3, 2, 5, 3, 3]), key=abs) == (T(2), T(5))


@memory_leak_decorator()
def test_minmax_key1():
    assert minmax(T('a'), T('b'), T('c'),
                  key=lambda x: x.value.upper()) == (T('a'), T('c'))


@memory_leak_decorator()
def test_minmax_key2():
    assert minmax(T((T(1), T(2))), T((T(2), T(3))), T((T(3), T(1))),
                  key=lambda x: x.value[1]) == (T((T(3), T(1))),
                                                T((T(2), T(3))))


@memory_leak_decorator()
def test_minmax_default1():
    assert minmax([], default=T(10)) == (T(10), T(10))


@memory_leak_decorator()
def test_minmax_stability1():
    assert minmax([T((T(1), T(5)))],
                  key=lambda x: x.value[0]) == (T((T(1), T(5))),
                                                T((T(1), T(5))))


@memory_leak_decorator()
def test_minmax_stability2():
    assert minmax(T((T(1), T(5))), T((T(1), T(1))),
                  key=lambda x: x.value[0]) == (T((T(1), T(5))),
                                                T((T(1), T(5))))


@memory_leak_decorator()
def test_minmax_stability3():
    assert minmax(T((T(1), T(5))), T((T(1), T(1))), T((T(1), T(2))),
                  key=lambda x: x.value[0]) == (T((T(1), T(5))),
                                                T((T(1), T(5))))


@memory_leak_decorator()
def test_minmax_stability4():
    assert minmax(T((T(1), T(5))), T((T(1), T(1))), T((T(1), T(2))),
                  T((T(1), T(3))),
                  key=lambda x: x.value[0]) == (T((T(1), T(5))),
                                                T((T(1), T(5))))


@memory_leak_decorator()
def test_minmax_stability5():
    assert minmax(T((T(5), T(5))), T((T(1), T(5))),
                  T((T(1), T(2))), T((T(1), T(3))),
                  key=lambda x: x.value[0]) == (T((T(1), T(5))),
                                                T((T(5), T(5))))


@memory_leak_decorator()
def test_minmax_stability6():
    assert minmax(T((T(5), T(5))), T((T(3), T(5))),
                  T((T(1), T(5))), T((T(1), T(3))),
                  key=lambda x: x.value[0]) == (T((T(1), T(5))),
                                                 T((T(5), T(5))))


@memory_leak_decorator()
def test_minmax_stability7():
    assert minmax(T((T(5), T(5))), T((T(3), T(5))),
                  T((T(4), T(5))), T((T(1), T(5))),
                  key=lambda x: x.value[0]) == (T((T(1), T(5))),
                                                T((T(5), T(5))))


@memory_leak_decorator(collect=True)
def test_minmax_failure1():
    # No args
    with pytest.raises(TypeError):
        minmax()


@memory_leak_decorator(collect=True)
def test_minmax_failure2():
    # empty sequence no default
    with pytest.raises(ValueError):
        minmax([])


@memory_leak_decorator(collect=True)
def test_minmax_failure3():
    # invalid kwarg
    with pytest.raises(TypeError):
        minmax(T(1), T(2), invalid_kw='a')


@memory_leak_decorator(collect=True)
def test_minmax_failure4():
    # default with multiple args
    with pytest.raises(TypeError):
        minmax(T(1), T(2), default=T(10))


@memory_leak_decorator(collect=True)
def test_minmax_failure5():
    # arg is not iterable
    with pytest.raises(TypeError):
        minmax(T(100))


@memory_leak_decorator(collect=True)
def test_minmax_failure6():
    # func fails on odd numbered arg
    with pytest.raises(TypeError):
        minmax(T(100), T('a'), key=lambda x: x.value + '')


@memory_leak_decorator(collect=True)
def test_minmax_failure7():
    # func fails on even numbered arg
    with pytest.raises(TypeError):
        minmax(T('a'), T(100), key=lambda x: x.value + '')


@pytest.mark.xfail(iteration_utilities.PY2, reason='cmp works on Python 2')
@memory_leak_decorator(collect=True)
def test_minmax_failure8():
    # unable to compare first and second
    with pytest.raises(TypeError):
        minmax(T(100), T('a'))


@pytest.mark.xfail(iteration_utilities.PY2, reason='cmp works on Python 2')
@memory_leak_decorator(collect=True)
def test_minmax_failure9():
    # unable to compare third and fourth
    with pytest.raises(TypeError):
        minmax(T(100), T(20), T(100), T('a'))


@pytest.mark.xfail(iteration_utilities.PY2, reason='cmp works on Python 2')
@memory_leak_decorator(collect=True)
def test_minmax_failure10():
    # unable to compare first and third
    with pytest.raises(TypeError):
        minmax(T(1), T(20), T('a'), T('c'))


@pytest.mark.xfail(iteration_utilities.PY2, reason='cmp works on Python 2')
@memory_leak_decorator(collect=True)
def test_minmax_failure11():
    # unable to compare second and fourth

    # This is tricky. The elements are explicitly chosen so that
    # 1 compares with 2 without error: 1 current min, 2 current max
    # 3 compares with 4: 3 < 4
    # 3 compares with 1: 1 still current minimum
    # 4 compares not with 2 because the first element is equal and then
    # the comparison with the second element throws and error because
    # str and int are not compareable.
    with pytest.raises(TypeError):
        minmax(T((100, 'a')), T((200, 10)), T((150, 'b')), T((200, 'd')))



# This module
import iteration_utilities

if iteration_utilities.PY2:
    from itertools import imap as map


class T(object):
    def __init__(self, value):
        self.value = value

    def _cmp_cls_and_value(self, other):
        if (type(self) != type(other) or
                type(self.value) != type(other.value)):
            raise TypeError('simulated failure.')

    # Misc
    def __hash__(self):
        return hash(self.value)

    def __bool__(self):
        return bool(self.value)

    def __nonzero__(self):
        return bool(self.value)

    def __len__(self):
        return len(self.value)

    def __repr__(self):
        return '{0.__class__.__name__}({0.value})'.format(self)

    # Mathematical
    def __add__(self, other):
        self._cmp_cls_and_value(other)
        return self.__class__(self.value + other.value)

    def __mul__(self, other):
        self._cmp_cls_and_value(other)
        return self.__class__(self.value * other.value)

    def __rtruediv__(self, other):
        return self.__class__(other / self.value)

    def __pow__(self, other):
        if isinstance(other, T):
            return self.__class__(self.value**other.value)
        else:
            return self.__class__(self.value**other)

    def __abs__(self):
        return self.__class__(abs(self.value))

    # Comparisons
    def __eq__(self, other):
        self._cmp_cls_and_value(other)
        return self.value == other.value

    def __lt__(self, other):
        self._cmp_cls_and_value(other)
        return self.value < other.value

    def __le__(self, other):
        self._cmp_cls_and_value(other)
        return self.value <= other.value

    def __gt__(self, other):
        self._cmp_cls_and_value(other)
        return self.value > other.value

    def __ge__(self, other):
        self._cmp_cls_and_value(other)
        return self.value >= other.value


def toT(iterable):
    return list(map(T, iterable))



# Built-ins
from __future__ import absolute_import, division, print_function
import itertools
import operator
import pickle

# 3rd party
import pytest

# This module
import iteration_utilities

# Test helper
from helper_leak import memory_leak_decorator
from helper_cls import T, toT


merge = iteration_utilities.merge


@memory_leak_decorator()
def test_merge_empty1():
    assert list(merge()) == []


@memory_leak_decorator()
def test_merge_empty2():
    assert list(merge([])) == []


@memory_leak_decorator()
def test_merge_empty3():
    assert list(merge([], (), {})) == []


@memory_leak_decorator()
def test_merge_normal1():
    for seq in itertools.permutations([[T(1)], [T(2)], [T(3)]]):
        assert list(merge(*seq)) == toT([1, 2, 3])


@memory_leak_decorator()
def test_merge_normal2():
    for seq in itertools.permutations([[T(1)], [T(2)], [T(3)], []]):
        assert list(merge(*seq)) == toT([1, 2, 3])


@memory_leak_decorator()
def test_merge_normal3():
    for seq in itertools.permutations([[T(1), T(3)], [T(2)], [T(4)]]):
        assert list(merge(*seq)) == toT([1, 2, 3, 4])


@memory_leak_decorator()
def test_merge_normal4():
    for seq in itertools.permutations([[T(1), T(3)], [T(0), T(2)], [T(4)]]):
        assert list(merge(*seq)) == toT([0, 1, 2, 3, 4])


@memory_leak_decorator()
def test_merge_normal5():
    perms = itertools.permutations([toT(range(5)),
                                    toT(range(3)),
                                    toT(range(4, 7))])
    for seq in perms:
        assert list(merge(*seq)) == toT([0, 0, 1, 1, 2, 2, 3, 4, 4, 5, 6])


@memory_leak_decorator()
def test_merge_stable1():
    # Stability tests (no use of T on purpose!)
    it = merge([1], [1.])
    item1 = next(it)
    assert isinstance(item1, int)
    item2 = next(it)
    assert isinstance(item2, float)


@memory_leak_decorator()
def test_merge_key1():
    # Key function tests
    seq = ([(T(1), T(0)), (T(2), T(0))], [(T(1), T(-1)), (T(2), T(-1))])
    assert (list(merge(*seq, key=operator.itemgetter(0))) ==
            [(T(1), T(0)), (T(1), T(-1)), (T(2), T(0)), (T(2), T(-1))])


@memory_leak_decorator()
def test_merge_reverse1():
    # Reverse test
    for seq in itertools.permutations([[T(1)], [T(2)], [T(3)]]):
        assert list(merge(*seq, reverse=True)) == toT([3, 2, 1])


@memory_leak_decorator()
def test_merge_reverse2():
    for seq in itertools.permutations([[T(1)], [T(2)], [T(3)], []]):
        assert list(merge(*seq, reverse=True)) == toT([3, 2, 1])


@memory_leak_decorator()
def test_merge_reverse3():
    for seq in itertools.permutations([[T(3), T(1)], [T(2)], [T(4)]]):
        assert list(merge(*seq, reverse=True)) == toT([4, 3, 2, 1])


@memory_leak_decorator()
def test_merge_reverse4():
    for seq in itertools.permutations([[T(3), T(1)], [T(2), T(0)], [T(4)]]):
        assert list(merge(*seq, reverse=True)) == toT([4, 3, 2, 1, 0])


@memory_leak_decorator()
def test_merge_keyreverse1():
    # Key+reverse function tests
    seq = ([(T(2), T(0)), (T(1), T(0))], [(T(2), T(-1)), (T(1), T(-1))])
    assert (list(merge(*seq, reverse=True, key=operator.itemgetter(0))) ==
            [(T(2), T(0)), (T(2), T(-1)), (T(1), T(0)), (T(1), T(-1))])


@memory_leak_decorator(collect=True)
def test_merge_failure1():
    # One iterable is not iterable
    with pytest.raises(TypeError):
        merge(T(10))


@memory_leak_decorator(collect=True)
def test_merge_failure2():
    # One iterable is not iterable
    with pytest.raises(TypeError):
        merge([T(10), T(20)], T(10))


@memory_leak_decorator(collect=True)
def test_merge_failure3():
    # Unexpected keyword argument
    with pytest.raises(TypeError):
        merge([T(10), T(20)], [T(20)], reverse=True, key=abs, wrongkwd=True)


@memory_leak_decorator(collect=True)
def test_merge_failure4():
    # Unexpected keyword argument
    with pytest.raises(TypeError):
        merge([T(10), T(20)], [T(20), T(30)], reverse=True, wrongkwd=True)


@memory_leak_decorator(collect=True)
def test_merge_failure5():
    # Unexpected keyword argument
    with pytest.raises(TypeError):
        merge([T(10), T(20)], [T(20), T(30)], key=abs, wrongkwd=True)


@memory_leak_decorator(collect=True)
def test_merge_failure6():
    # Unexpected keyword argument
    with pytest.raises(TypeError):
        merge([T(10), T(20)], [T(20), T(30)], wrongkwd=True)


@memory_leak_decorator(collect=True)
def test_merge_failure7():
    # Key function fails
    with pytest.raises(TypeError):
        list(merge([T(2), (T(2), T(0))], [(T(1), T(2)), (T(1), T(3))],
                   key=operator.itemgetter(0)))


@memory_leak_decorator(collect=True)
def test_merge_failure8():
    # Key function fails
    with pytest.raises(TypeError):
        list(merge([(T(2), T(0)), T(2)], [(T(1), T(2)), (T(1), T(3))],
                   key=operator.itemgetter(0)))


@pytest.mark.xfail(iteration_utilities.PY2, reason='cmp works on Python 2')
@memory_leak_decorator(collect=True)
def test_merge_failure9():
    # comparison fails
    with pytest.raises(TypeError):
        list(merge([T('a'), T('b')], [T(2), T(3)]))


@pytest.mark.xfail(iteration_utilities.PY2, reason='cmp works on Python 2')
@memory_leak_decorator(collect=True)
def test_merge_failure10():
    # comparison fails
    with pytest.raises(TypeError):
        list(merge([T(1), T('b')], [T(2), T(3)]))


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_merge_pickle1():
    mge = merge([T(0)], [T(1), T(2)], [T(2)])
    assert next(mge) == T(0)
    x = pickle.dumps(mge)
    assert list(pickle.loads(x)) == toT([1, 2, 2])


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_merge_pickle2():
    mge = merge([T(1), T(2)], [T(0)], [T(-2)], key=abs)
    assert next(mge) == T(0)
    x = pickle.dumps(mge)
    assert list(pickle.loads(x)) == toT([1, 2, -2])


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_merge_pickle3():
    mge = merge([T(2), T(1)], [T(0)], [T(3)], reverse=True)
    assert next(mge) == T(3)
    x = pickle.dumps(mge)
    assert list(pickle.loads(x)) == toT([2, 1, 0])



# Built-ins
from __future__ import absolute_import, division, print_function

# 3rd party
import pytest

# This module
import iteration_utilities

# Test helper
from helper_leak import memory_leak_decorator
from helper_cls import T


all_equal = iteration_utilities.all_equal


@memory_leak_decorator()
def test_all_equal_empty1():
    assert all_equal([])


@memory_leak_decorator()
def test_all_equal_normal1():
    assert all_equal([T(1), T(1), T(1)])


@memory_leak_decorator()
def test_all_equal_normal2():
    assert not all_equal([T(1), T(1), T(2)])


@memory_leak_decorator(collect=True)
def test_all_equal_failure1():
    # not iterable
    with pytest.raises(TypeError):
        all_equal(T(1))


@memory_leak_decorator(collect=True)
def test_all_equal_failure2():
    # comparison fail
    with pytest.raises(TypeError):
        all_equal([T(1), T('a')])



# Built-ins
from __future__ import absolute_import, division, print_function
from functools import partial
from operator import eq
import pickle

# 3rd party
import pytest

# This module
import iteration_utilities

# Test helper
from helper_leak import memory_leak_decorator
from helper_cls import T, toT


split = iteration_utilities.split


equalsthreeT = partial(eq, T(3))


@memory_leak_decorator()
def test_split_empty1():
    assert list(split([], lambda x: False)) == []


@memory_leak_decorator()
def test_split_normal1():
    assert list(split([T(1), T(2), T(3)],
                      lambda x: x.value == 2)) == [[T(1)], [T(3)]]


@memory_leak_decorator()
def test_split_normal2():
    assert list(split([T(1), T(2), T(3)],
                      lambda x: x.value == 3)) == [toT([1, 2])]


@memory_leak_decorator()
def test_split_keep1():
    assert list(split([T(1), T(2), T(3)], lambda x: x.value == 2,
                      keep=True)) == [[T(1)], [T(2)], [T(3)]]


@memory_leak_decorator()
def test_split_keep2():
    assert list(split([T(1), T(2), T(3)], lambda x: x.value == 3,
                      keep=True)) == [[T(1), T(2)], [T(3)]]


@memory_leak_decorator()
def test_split_maxsplit1():
    assert list(split([T(1), T(2), T(3), T(4), T(5)],
                      lambda x: x.value % 2 == 0,
                      maxsplit=1)) == [[T(1)], [T(3), T(4), T(5)]]


@memory_leak_decorator()
def test_split_maxsplit2():
    assert list(split([T(1), T(2), T(3), T(4), T(5)],
                      lambda x: x.value % 2 == 0,
                      maxsplit=2)) == [[T(1)], [T(3)], [T(5)]]


@memory_leak_decorator()
def test_split_eq1():
    assert list(split([T(1), T(2), T(3), T(2), T(5)], T(2),
                      eq=True)) == [[T(1)], [T(3)], [T(5)]]


@memory_leak_decorator(collect=True)
def test_split_failure1():
    # not iterable
    with pytest.raises(TypeError):
        split(T(1), lambda x: False)


@memory_leak_decorator(collect=True)
def test_split_failure2():
    # func fails
    with pytest.raises(TypeError):
        list(split([T(1), T(2), T(3)], lambda x: T(x.value + 'a')))


@memory_leak_decorator(collect=True)
def test_split_failure3():
    # cmp fails
    with pytest.raises(TypeError):
        list(split([T(1), T(2), T(3)], T('a'), eq=True))


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_split_pickle1():
    l = [T(1), T(2), T(3), T(4), T(5), T(3), T(7), T(8)]
    spl = split(l, equalsthreeT)
    x = pickle.dumps(spl)
    assert list(pickle.loads(x)) == [[T(1), T(2)], [T(4), T(5)], [T(7), T(8)]]


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_split_pickle2():
    l = [T(1), T(2), T(3), T(4), T(5), T(3), T(7), T(8)]
    spl = split(l, equalsthreeT)
    assert next(spl) == toT([1, 2])
    x = pickle.dumps(spl)
    assert list(pickle.loads(x)) == [toT([4, 5]), toT([7, 8])]


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_split_pickle3():
    l = [T(1), T(2), T(3), T(4), T(5), T(3), T(7), T(8)]
    spl = split(l, equalsthreeT, keep=True)
    assert next(spl) == toT([1, 2])
    x = pickle.dumps(spl)
    assert list(pickle.loads(x)) == [toT(i)
                                     for i in [[3], [4, 5], [3], [7, 8]]]


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_split_pickle4():
    l = [T(1), T(2), T(3), T(4), T(5), T(3), T(7), T(8)]
    spl = split(l, equalsthreeT, maxsplit=1)
    assert next(spl) == toT([1, 2])
    x = pickle.dumps(spl)
    assert list(pickle.loads(x)) == [toT([4, 5, 3, 7, 8])]


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_split_pickle5():
    l = [T(1), T(2), T(3), T(4), T(5), T(3), T(7), T(8)]
    spl = split(l, T(3), eq=True)
    assert next(spl) == toT([1, 2])
    x = pickle.dumps(spl)
    assert list(pickle.loads(x)) == [toT([4, 5]), toT([7, 8])]



# Built-ins
from __future__ import absolute_import, division, print_function
from itertools import tee

# 3rd party
import pytest

# This module
import iteration_utilities

# Test helper
from helper_doctest import doctest_module_no_failure


def test_doctests():
    doctest_module_no_failure(iteration_utilities._recipes._core)
    doctest_module_no_failure(iteration_utilities._recipes._additional)
    doctest_module_no_failure(iteration_utilities._helpers._performance)


def test_exceptions():
    # old-style classes don't have the subclasses special member.
    if iteration_utilities.PY2:
        class A:
            pass
        with pytest.raises(TypeError):
            list(iteration_utilities.itersubclasses(A))

    # Random product doesn't work with empty iterables
    with pytest.raises(IndexError):
        iteration_utilities.random_product([])

    # There is no element 10 in the tee object so this will raise the
    # Exception.
    t1, t2 = tee([1, 2, 3, 4, 5])
    with pytest.raises(IndexError):
        iteration_utilities.tee_lookahead(t1, 10)

    # Missing idx or start/stop in replace/remove/getitem
    with pytest.raises(TypeError):
        iteration_utilities.replace([1, 2, 3], 5)
    with pytest.raises(TypeError):
        iteration_utilities.remove([1, 2, 3])
    with pytest.raises(TypeError):
        iteration_utilities.getitem([1, 2, 3])
    # Stop smaller than start in replace/remove
    with pytest.raises(ValueError):
        iteration_utilities.replace(range(10), 5, start=7, stop=5)
    with pytest.raises(ValueError):
        iteration_utilities.remove(range(10), start=7, stop=5)
    # idx smaller than -1 in getitem
    with pytest.raises(ValueError):
        iteration_utilities.getitem(range(10), (4, 2, -3, 9))


def test_empty_input():
    empty = []

    assert iteration_utilities.all_isinstance(empty, float)

    assert not iteration_utilities.any_isinstance(empty, float)

    assert iteration_utilities.consume(empty, 2) is None

    assert list(iteration_utilities.deepflatten(empty)) == []

    assert iteration_utilities.dotproduct(empty, empty) == 0

    assert list(iteration_utilities.flatten(empty)) == []

    assert list(iteration_utilities.getitem(
        range(10), empty)) == []

    x, y = iteration_utilities.ipartition(empty, lambda x: x)
    assert list(x) == [] and list(y) == []

    # no need to test iter_subclasses here

    assert list(iteration_utilities.ncycles(empty, 10)) == []

    assert list(iteration_utilities.powerset(empty)) == [()]

    assert iteration_utilities.random_combination(empty, 0) == ()
    assert iteration_utilities.random_combination(empty, 0, True) == ()

    assert iteration_utilities.random_permutation(empty, 0) == ()

    assert list(iteration_utilities.remove(
        range(10), empty)) == list(range(10))

    assert list(iteration_utilities.replace(
        range(10), 20, empty)) == list(range(10))

    # no need to test repeatfunc here

    # no need to test tabulate here

    assert list(iteration_utilities.tail(empty, 2)) == []

    # no need to test tee_lookahead here



# Built-ins
from __future__ import absolute_import, division, print_function
import pickle

# 3rd party
import pytest

# This module
import iteration_utilities

# Test helper
from helper_leak import memory_leak_decorator
from helper_cls import T


flip = iteration_utilities.flip


@memory_leak_decorator()
def test_flip_normal1():
    assert not flip(isinstance)(float, 10)


@memory_leak_decorator()
def test_flip_normal2():
    assert flip(isinstance)(int, 10)


@memory_leak_decorator()
def test_flip_args0():
    def func():
        return ()
    assert flip(func)() == ()


@memory_leak_decorator()
def test_flip_args1():
    def func(a):
        return (a, )
    assert flip(func)(T(10)) == (T(10), )


@memory_leak_decorator()
def test_flip_args2():
    def func(a, b):
        return a, b
    assert flip(func)(T(1), T(2)) == (T(2), T(1))


@memory_leak_decorator()
def test_flip_args3():
    def func(a, b, c):
        return a, b, c
    assert flip(func)(T(1), T(2), T(3)) == (T(3), T(2), T(1))


@memory_leak_decorator(collect=True)
def test_flip_failure1():
    with pytest.raises(TypeError):
        flip(isinstance)(10, float)


@memory_leak_decorator(offset=1)
def test_flip_pickle1():
    x = pickle.dumps(flip(isinstance))
    assert pickle.loads(x)(float, 10.)
    assert pickle.loads(x)(int, 10)
    assert not pickle.loads(x)(float, 10)



# Built-ins
from __future__ import absolute_import, division, print_function

# 3rd party
import pytest

# This module
import iteration_utilities

# Test helper
from helper_leak import memory_leak_decorator
from helper_cls import T


all_distinct = iteration_utilities.all_distinct


@memory_leak_decorator()
def test_alldistinct_empty1():
    assert all_distinct([])


@memory_leak_decorator()
def test_alldistinct_normal1():
    assert all_distinct([T(1), T(2), T(3)])


@memory_leak_decorator()
def test_alldistinct_normal2():
    assert not all_distinct([T(1), T(1), T(1)])


@memory_leak_decorator()
def test_alldistinct_unhashable1():
    assert all_distinct([{T('a'): T(1)}, {T('a'): T(2)}])


@memory_leak_decorator()
def test_alldistinct_unhashable2():
    assert not all_distinct([{T('a'): T(1)}, {T('a'): T(1)}])


@memory_leak_decorator(collect=True)
def test_alldistinct_failure1():
    # iterable is not iterable
    with pytest.raises(TypeError):
        all_distinct(T(1))



# Built-ins
from __future__ import absolute_import, division, print_function
import pickle

# 3rd party
import pytest

# This module
import iteration_utilities

# Test helper
from helper_leak import memory_leak_decorator
from helper_cls import T


Seen = iteration_utilities.Seen


@memory_leak_decorator()
def test_seen_equality0():
    assert Seen() == Seen()


@memory_leak_decorator()
def test_seen_equality1():
    assert Seen({T(1), T(2)}) == Seen({T(1), T(2)})


@memory_leak_decorator()
def test_seen_equality2():
    assert not Seen({T(1), T(2), T(3)}) == Seen({T(1), T(2)})


@memory_leak_decorator()
def test_seen_equality3():
    assert Seen({T(1)}, [T([0, 0])]) == Seen({T(1)}, [T([0, 0])])


@memory_leak_decorator()
def test_seen_equality4():
    assert not Seen({T(1)}, [T([0, 0])]) == Seen({T(1)}, [T([0, 1])])


@memory_leak_decorator()
def test_seen_equality5():
    assert not Seen({T(1)}, [T([0, 0]), T([1, 0])]) == Seen({T(1)}, [T([0, 1])])


@memory_leak_decorator()
def test_seen_equality6():
    assert Seen(set()) == Seen(set(), [])


@memory_leak_decorator()
def test_seen_equality7():
    assert Seen(set(), []) == Seen(set())


@memory_leak_decorator()
def test_seen_nequality0():
    assert not Seen() != Seen()


@memory_leak_decorator()
def test_seen_nequality1():
    assert not Seen({T(1), T(2)}) != Seen({T(1), T(2)})


@memory_leak_decorator()
def test_seen_nequality2():
    assert Seen({T(1), T(2), T(3)}) != Seen({T(1), T(2)})


@memory_leak_decorator()
def test_seen_nequality3():
    assert not Seen({T(1)}, [T([0, 0])]) != Seen({T(1)}, [T([0, 0])])


@memory_leak_decorator()
def test_seen_nequality4():
    assert Seen({T(1)}, [T([0, 0])]) != Seen({T(1)}, [T([0, 1])])


@memory_leak_decorator()
def test_seen_nequality5():
    assert Seen({T(1)}, [T([0, 0]), [T(1), T(0)]]) != Seen({T(1)}, [T([0, 0])])


@memory_leak_decorator()
def test_seen_nequality6():
    assert not Seen(set()) != Seen(set(), [])


@memory_leak_decorator()
def test_seen_nequality7():
    assert not Seen(set(), []) != Seen(set())


@memory_leak_decorator()
def test_seen_len0():
    assert not len(Seen())
    assert len(Seen({T(1), T(2), T(3)})) == 3
    assert len(Seen(seenlist=[[T(0), T(0)], [T(1), T(1)], [T(2), T(2)]])) == 3
    assert len(Seen({T(1), T(2), T(3)},
                    seenlist=[[T(0), T(0)], [T(1), T(1)], [T(2), T(2)]])) == 6


@memory_leak_decorator()
def test_seen_contains0():
    x = Seen()
    assert T(1) not in x
    assert x == Seen(set())
    assert T([0, 0]) not in x
    assert x == Seen(set())


@memory_leak_decorator()
def test_seen_containsadd0():
    x = Seen()
    assert not x.contains_add(T(1))
    assert not x.contains_add(T([0, 0]))
    assert T(1) in x
    assert T([0, 0]) in x
    assert x == Seen({T(1)}, [T([0, 0])])


# Pickle tests and most failure tests are implemented implicitly as part of
# unique_everseen, duplicates, all_distinct so there should be no need to
# repeat these here. But if "Seen" is expanded these should be included!!!



# Built-ins
from __future__ import absolute_import, division, print_function
import pickle

# 3rd party
import pytest

# This module
import iteration_utilities

# Test helper
from helper_leak import memory_leak_decorator
from helper_cls import T, toT

string_types = basestring if iteration_utilities.PY2 else str


deepflatten = iteration_utilities.deepflatten


@memory_leak_decorator()
def test_deepflatten_empty1():
    assert list(deepflatten([])) == []


@memory_leak_decorator()
def test_deepflatten_normal1():
    assert list(deepflatten([T(1), T(2), T(3)])) == [T(1), T(2), T(3)]


@memory_leak_decorator()
def test_deepflatten_normal2():
    assert list(deepflatten([[T(1)], T(2), [[T(3)]]])) == toT([1, 2, 3])


@memory_leak_decorator()
def test_deepflatten_normal3():
    # really deeply nested thingy
    assert list(deepflatten([[[[[[[[[[[T(5), T(4), T(3), T(2), T(1), T(0)]]]]],
                               map(T, range(3))]]],
                            (T(i) for i in range(5))]]])
                ) == toT([5, 4, 3, 2, 1, 0, 0, 1, 2, 0, 1, 2, 3, 4])


@memory_leak_decorator()
def test_deepflatten_normal4():
    # really deeply nested thingy with types
    assert list(deepflatten([[[[[[[[[[[T(5), T(4), T(3), T(2), T(1), T(0)]]]]],
                               [T(0), T(1), T(2)]]]],
                            [T(0), T(1), T(2), T(3), T(4)]]]], types=list)
                ) == toT([5, 4, 3, 2, 1, 0, 0, 1, 2, 0, 1, 2, 3, 4])


@memory_leak_decorator()
def test_deepflatten_depth1():
    assert list(deepflatten([T(1), T(2), T(3)], 1)) == toT([1, 2, 3])


@memory_leak_decorator()
def test_deepflatten_depth2():
    assert list(deepflatten([[T(1)], T(2), [[T(3)]]],
                            1)) == [T(1), T(2), [T(3)]]


@memory_leak_decorator()
def test_deepflatten_types1():
    assert list(deepflatten([[T(1)], T(2), [[T(3)]]],
                            types=list)) == toT([1, 2, 3])


@memory_leak_decorator()
def test_deepflatten_types2():
    assert list(deepflatten([[T(1)], T(2), [[T(3)]]],
                            types=tuple)) == [[T(1)], T(2), [[T(3)]]]


@memory_leak_decorator()
def test_deepflatten_types3():
    assert list(deepflatten([[T(1)], T(2), ([T(3)], )],
                            types=(list, tuple))) == toT([1, 2, 3])


@memory_leak_decorator()
def test_deepflatten_ignore1():
    assert list(deepflatten([[T(1)], T(2), [[T(3), 'abc']]],
                            ignore=string_types)) == [T(1), T(2), T(3), 'abc']


@memory_leak_decorator()
def test_deepflatten_ignore2():
    assert list(deepflatten([[T(1)], T(2), ([T(3), 'abc'], )],
                            ignore=(tuple, string_types))
                ) == [T(1), T(2), ([T(3), 'abc'], )]


@memory_leak_decorator(collect=True)
def test_deepflatten_failure1():
    with pytest.raises(TypeError):
        list(deepflatten([T(1), T(2), T(3)], None, T('a')))


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_deepflatten_pickle1():
    dpflt = deepflatten([[T(1)], [T(2)], [T(3)], [T(4)]])
    assert next(dpflt) == T(1)
    x = pickle.dumps(dpflt)
    assert list(pickle.loads(x)) == toT([2, 3, 4])



# Built-ins
from __future__ import absolute_import, division, print_function

# 3rd party

# This module
import iteration_utilities

# Test helper


def test_other_c_funcs():
    assert iteration_utilities.return_True()
    assert not iteration_utilities.return_False()
    assert iteration_utilities.return_None() is None
    assert iteration_utilities.return_identity(1) == 1
    assert iteration_utilities.return_first_arg(1, 2, 3) == 1
    assert iteration_utilities.return_called(int) == 0

    assert iteration_utilities.square(2) == 4
    assert iteration_utilities.reciprocal(2) == 0.5

    assert iteration_utilities.is_None(None)
    assert not iteration_utilities.is_None(False)

    assert not iteration_utilities.is_not_None(None)
    assert iteration_utilities.is_not_None(False)

    assert iteration_utilities.is_even(2)
    assert not iteration_utilities.is_even(1)

    assert iteration_utilities.is_odd(1)
    assert not iteration_utilities.is_odd(2)

    assert not iteration_utilities.is_iterable(1)
    assert iteration_utilities.is_iterable([1])



# Built-ins
from __future__ import absolute_import, division, print_function

# 3rd party

# This module
import iteration_utilities

# Test helper
from helper_doctest import doctest_module_no_failure


def test_doctests():
    # classes are added to the main module code. :-)
    doctest_module_no_failure(iteration_utilities)
    doctest_module_no_failure(iteration_utilities._cfuncs)



# Built-ins
from __future__ import absolute_import, division, print_function

# 3rd party
import pytest

# This module
import iteration_utilities

# Test helper
from helper_leak import memory_leak_decorator
from helper_cls import T


iter_except = iteration_utilities.iter_except


@memory_leak_decorator()
def test_iterexcept_normal1():
    assert list(iter_except(({T('a'): T(10)}).popitem, KeyError
                            )) == [(T('a'), T(10))]


@memory_leak_decorator()
def test_iterexcept_first():
    d = {}

    def insert():
        d[T('a')] = T(10)

    assert list(iter_except(d.popitem, KeyError, insert
                            )) == [None, (T('a'), T(10))]


@memory_leak_decorator(collect=True)
def test_iterexcept_failure1():
    # wrong exception
    with pytest.raises(KeyError):
        list(iter_except(({T('a'): T(10)}).popitem, ValueError))


@memory_leak_decorator(offset=1)
def test_iterexcept_pickle1():
    # TODO: Implement some tests for pickle!
    pass



# Built-ins
from __future__ import absolute_import, division, print_function
import pickle

# 3rd party
import pytest

# This module
import iteration_utilities

# Test helper
from helper_leak import memory_leak_decorator
from helper_cls import T


applyfunc = iteration_utilities.applyfunc
getitem = iteration_utilities.getitem


@memory_leak_decorator()
def test_applyfunc_normal1():
    assert list(getitem(applyfunc(lambda x: x**T(2), T(2)),
                        stop=3)) == [T(4), T(16), T(256)]


@memory_leak_decorator()
def test_applyfunc_normal2():
    assert list(getitem(applyfunc(lambda x: x, T(2)),
                        stop=3)) == [T(2), T(2), T(2)]


@memory_leak_decorator(collect=True)
def test_applyfunc_failure1():
    with pytest.raises(TypeError):
        list(getitem(applyfunc(lambda x: x**T(2), T('a')), stop=3))


@pytest.mark.xfail(iteration_utilities.PY2, reason='pickle does not work on Python 2')
@memory_leak_decorator(offset=1)
def test_applyfunc_pickle1():
    apf = applyfunc(iteration_utilities.square, T(2))
    assert next(apf) == T(4)
    x = pickle.dumps(apf)
    assert next(pickle.loads(x)) == T(16)
